<?php

namespace Users;

use Router;

function register_routes(Router $r): Router
{
	$r->form('/login', 'Users\login');
	$r->get('/logout', 'Users\logout');
	$r->form('/register', 'Users\register');
	$r->form('/lostpassword', 'Users\lostpassword');
	$r->form('/changepassword', 'Users\changepassword');
	$r->form('/verify', 'Users\verify');
	$r->form('/settings', 'Users\settings');
	return $r;
}

/**
 * Displays the login page, and processes login requests.
 */
function login()
{
	global $auth;

	if (is_post()) {
		$form = validate($_POST, [
			'username' => ['length:3-18', 'alpha-spaces'],
			'password' => ['length:6-255'],
			'remember' => ['bool']
		]);

		if (!$form['valid']) {
			exit(ul_from_validate_errors($form['errors']));
		}

		$form = $form['data'];
		$row = get_user($form['username']);
		if ($row === false || !$auth->login($form['username'], $form['password']))
			die("Invalid username or password. Please go back and try again.");

		$expiretime = $form['remember'] ? time() + 31536000 : 0;
		$rememberme = $form['remember'] ? 1 : 0;
		$cookie = implode(' ', [$row['id'], $row['username'], $row['password'], $rememberme]);

        set_cookie("dkgame", $cookie, $expiretime);
        redirect('/');
	}

    return display(render('login'), 'Log In', true, false, false);
}

/**
 * Delete the current cookie and redirect to home.
 */
function logout()
{
	global $auth;
	$auth->logout();
    set_cookie("dkgame", "", -3600);
    redirect('/login');
}

/**
 * Register a new account.
 */
function register()
{
    global $controlrow;

    if (isset($_POST["submit"])) {
		$form = validate($_POST, [
			'username'         => ['length:3-18', 'alpha-spaces', 'unique:users,username'],
			'email'            => ['email', 'unique:users,email'],
			'confirm_email'    => ['confirm'],
			'password'         => ['length:6-255'],
			'confirm_password' => ['confirm'],
			'charclass'        => ['in:1,2,3']
		]);

		if (!$form['valid']) {
			$err = ul_from_validate_errors($form['errors']);
			$page = "The following error(s) occurred when your account was being made:<br><span style=\"color:red;\">$err</span><br>Please go back and try again.";
		} else {
			$form = $form['data'];
			$password = password_hash($form['password'], PASSWORD_ARGON2ID);
			$token = ($controlrow['verifyemail'] == true) ? token(8) : 'g2g';
			db()->query('INSERT INTO users (verify, username, password, email, charclass) VALUES (?, ?, ?, ?, ?)', [
				$token, $form['username'], $password, $form['email'], $form['charclass']
			]);

			if ($controlrow['verifyemail'] == true) {
				if (sendregmail($form['email'], $token)) {
                    $page = "Your account was created successfully.<br><br>You should receive an Account Verification email shortly. You will need the verification code contained in that email before you are allowed to log in. Once you have received the email, please visit the <a href=\"users.php?do=verify\">Verification Page</a> to enter your code and start playing.";
                } else {
                    $page = "Your account was created successfully.<br><br>However, there was a problem sending your verification email. Please check with the game administrator to help resolve this problem.";
                }
			} else {
				$page = "Your account was created succesfully.<br><br>You may now continue to the <a href=\"/login\">Login Page</a> and continue playing ".$controlrow["gamename"]."!";
			}
		}
    } else {
        if ($controlrow["verifyemail"] == true) {
            $controlrow["verifytext"] = "<br><span class=\"small\">A verification code will be sent to the address above, and you will not be able to log in without first entering the code. Please be sure to enter your correct email address.</span>";
        } else {
            $controlrow["verifytext"] = "";
        }

		$page = render('register', ['controlrow' => $controlrow]);
    }

    return display($page, 'Register', true, false, false);
}

function verify()
{
    if (isset($_POST["submit"])) {
        $u = trim($_POST['username'] ?? '');
		$e = trim($_POST['email'] ?? '');
		$t = trim($_POST['token'] ?? '');

		$query = db()->query('SELECT id FROM users WHERE username=? AND email=? AND verify=? LIMIT 1;', [$u, $e, $t]);
        if ($query === false) exit('Verification failed. Go back, double-check your details, and try again.');

		db()->query("UPDATE users SET verify='g2g' WHERE username=?;", [$u]);

        return display("Your account was verified successfully.<br><br>You may now continue to the <a href=\"/login\">Login Page</a> and start playing the game.<br><br>Thanks for playing!","Verify Email",false,false,false);
    }

    return display(render('verify'), "Verify Email", true, false, false);
}

function lostpassword()
{
    if (isset($_POST["submit"])) {
        $e = trim($_POST['email'] ?? '');

        if (!db()->exists('users', 'email', $e)) exit("No account with that email address.");

		$newpass = token(16);
		$hashed = password_hash($newpass, PASSWORD_ARGON2ID);

		db()->query('UPDATE users SET password=? WHERE email=?;', [$hashed, $e]);

        if (sendpassemail($e, $newpass)) {
            return display("Your new password was emailed to the address you provided.<br><br>Once you receive it, you may <a href=\"/login\">Log In</a> and continue playing.<br><br>Thank you.","Lost Password",false,false,false);
        } else {
            return display("There was an error sending your new password.<br><br>Please check with the game administrator for more information.<br><br>We apologize for the inconvience.","Lost Password",false,false,false);
        }
    }

    return display(render('lostpassword'), "Lost Password", true, false, false);
}

function changepassword()
{
    if (isset($_POST["submit"])) {
        $u = trim($_POST['username'] ?? '');
		$p = $_POST['password'] ?? '';
		$np = $_POST['new_password'] ?? '';
		$np2 = $_POST['new_password2'] ?? '';

		$user = db()->query("SELECT password FROM users WHERE username=? LIMIT 1;", [$u]);
		$user = $user->fetchArray(SQLITE3_ASSOC);
        if ($user === false) exit("No account with that username.");

        if (!password_verify($p, $user['password'])) exit("The old password you provided was incorrect.");

		if (empty($np) || strlen($np) < 6) {
			$errors[] = 'New password is required and must be at least 6 characters long.';
		}

		if ($np2 !== $np) {
			$errors[] = 'Verify New Password must match.';
		}

        $realnewpass = password_hash($np, PASSWORD_ARGON2ID);
		db()->query('UPDATE users SET password=? WHERE username=?;', [$realnewpass, $u]);

		set_cookie('dkgame', '', -3600);

        return display("Your password was changed successfully.<br><br>You have been logged out of the game to avoid errors.<br><br>Please <a href=\"/login\">log back in</a> to continue playing.","Change Password",false,false,false);
    }

    return display(render('changepassword'), "Change Password", true, false, false);
}

function settings()
{
	if (is_post()) {
		$form = validate($_POST, [
			'game_skin' => ['in:0,1']
		]);
		if (!$form['valid']) exit(ul_from_validate_errors($form['errors']));
		$form = $form['data'];

		user()->game_skin = $form['game_skin'];
		user()->save();

		$alert = '<div class="alert">Settings updated</div>';
		return display($alert . render('settings'), "Account Settings");
	}

	return display(render('settings'), "Account Settings");
}

function sendpassemail($emailaddress, $password)
{
    global $controlrow;
    extract($controlrow);

	$email = <<<HTML
	You or someone using your email address submitted a Lost Password application on the $gamename server, located at $gameurl.

	We have issued you a new password so you can log back into the game.

	Your new password is: $password

	Thanks for playing.
	HTML;

	return send_email($emailaddress, "$gamename Lost Password", $email);
}

function sendregmail($emailaddress, $vercode)
{
    global $controlrow;
    extract($controlrow);
    $verurl = $gameurl . "/verify";

	$email = <<<HTML
	You or someone using your email address recently signed up for an account on the $gamename server, located at $gameurl.

	This email is sent to verify your registration email. In order to begin using your account, you must verify your email address.
	Please visit the Verification Page ($verurl) and enter the code below to activate your account.
	Verification code: $vercode

	If you were not the person who signed up for the game, please disregard this message. You will not be emailed again.
	HTML;

	return send_email($emailaddress, "$gamename Account Verification", $email);
}
